@@ -297,119 +297,150 @@ datetime_from_filetime (guint64 ft)
 	return res;
 }
 
+static char *
+get_dirent_name (gchar const *data, guint16 name_len)
+{
+	gchar const *end;
+
+	if (name_len <= 0 || name_len > DIRENT_MAX_NAME_SIZE)
+		return NULL;
+
+	/* !#%!@$#^
+	 * Sometimes, rarely, people store the stream name as ascii
+	 * rather than utf16.  Do a validation first just in case.
+	 */
+	if (!g_utf8_validate (data, name_len, &end) ||
+	    end - data + 1 != name_len) {
+		gunichar2 uni_name[DIRENT_MAX_NAME_SIZE + 1];
+		int i;
+
+		/* be wary about endianness */
+		for (i = 0 ; i < name_len ; i += 2)
+			uni_name[i/2] = GSF_LE_GET_GUINT16 (data + i);
+		uni_name[i/2] = 0;
+		return g_utf16_to_utf8 (uni_name, -1, NULL, NULL, NULL);
+	} else {
+		return g_strndup (data, end - data + 1);
+	}
+}
+
+
+typedef struct {
+	MSOleDirent *parent;
+	guint32 entry;
+} DirentEntry;
+
+static void
+add_todo (GArray *todo, MSOleDirent *parent, guint32 entry)
+{
+	size_t i = todo->len;
+	g_array_set_size (todo, i + 1);
+	g_array_index (todo, DirentEntry, i).parent = parent;
+	g_array_index (todo, DirentEntry, i).entry = entry;
+}
 
 /* Parse dirent number @entry and recursively handle its siblings and children.
  * parent is optional. */
-static MSOleDirent *
+static void
 ole_dirent_new (GsfInfileMSOle *ole, guint32 entry, MSOleDirent *parent,
 		GByteArray *seen_before)
 {
-	MSOleDirent *dirent;
-	guint32 block, next, prev, child, size;
-	guint8 const *data;
-	guint8 type;
-	guint16 name_len;
-	guint64 ft;
+	GArray *todo = g_array_new (FALSE, FALSE, sizeof (DirentEntry));
+	add_todo (todo, parent, entry);
 
-	if (entry >= DIRENT_MAGIC_END)
-		return NULL;
+	while (todo->len) {
+		parent = g_array_index (todo, DirentEntry, todo->len - 1).parent;
+		entry = g_array_index (todo, DirentEntry, todo->len - 1).entry;
+		g_array_set_size (todo, todo->len - 1);
 
-	g_return_val_if_fail (entry <= G_MAXUINT / DIRENT_SIZE, NULL);
-	g_return_val_if_fail (entry < seen_before->len, NULL);
+		if (entry >= DIRENT_MAGIC_END)
+			continue;
 
-	block = OLE_BIG_BLOCK (entry * DIRENT_SIZE, ole);
-	g_return_val_if_fail (block < ole->bat.num_blocks, NULL);
+		if (entry >= G_MAXUINT / DIRENT_SIZE ||
+		    entry >= seen_before->len ||
+		    seen_before->data[entry]) {
+		fail:
+			g_printerr ("Problem with directory entry %d\n", entry);
+			continue;
+		}
 
-	g_return_val_if_fail (!seen_before->data[entry], NULL);
-	seen_before->data[entry] = TRUE;
+		guint32 block = OLE_BIG_BLOCK (entry * DIRENT_SIZE, ole);
+		if (block >= ole->bat.num_blocks)
+			goto fail;
 
-	data = ole_get_block (ole, ole->bat.block [block], NULL);
-	if (data == NULL)
-		return NULL;
-	data += (DIRENT_SIZE * entry) % ole->info->bb.size;
+		guint8 const *data = ole_get_block (ole, ole->bat.block [block], NULL);
+		if (data == NULL)
+			goto fail;
+		data += (DIRENT_SIZE * entry) % ole->info->bb.size;
 
-	type = GSF_LE_GET_GUINT8 (data + DIRENT_TYPE);
-	if (type != DIRENT_TYPE_DIR &&
-	    type != DIRENT_TYPE_FILE &&
-	    type != DIRENT_TYPE_ROOTDIR) {
-		g_warning ("Unknown stream type 0x%x", type);
-		return NULL;
-	}
-	if (!parent && type != DIRENT_TYPE_ROOTDIR) {
-		/* See bug 346118.  */
-		g_warning ("Root directory is not marked as such.");
-		type = DIRENT_TYPE_ROOTDIR;
-	}
+		guint8 type = GSF_LE_GET_GUINT8 (data + DIRENT_TYPE);
+		if (type != DIRENT_TYPE_DIR &&
+		    type != DIRENT_TYPE_FILE &&
+		    type != DIRENT_TYPE_ROOTDIR) {
+			g_warning ("Unknown stream type 0x%x", type);
+			continue;
+		}
 
-	/* It looks like directory (and root directory) sizes are sometimes bogus */
-	size = GSF_LE_GET_GUINT32 (data + DIRENT_FILE_SIZE);
-	g_return_val_if_fail (type == DIRENT_TYPE_DIR || type == DIRENT_TYPE_ROOTDIR ||
-			      size <= (guint32)ole->input->size, NULL);
+		if (!parent && type != DIRENT_TYPE_ROOTDIR) {
+			/* See bug 346118.  */
+			g_warning ("Root directory is not marked as such.");
+			type = DIRENT_TYPE_ROOTDIR;
+		}
 
-	ft = GSF_LE_GET_GUINT64 (data + DIRENT_MODIFY_TIME);
+		/* It looks like directory (and root directory) sizes are sometimes bogus */
+		guint32 size = GSF_LE_GET_GUINT32 (data + DIRENT_FILE_SIZE);
+		if (type != DIRENT_TYPE_DIR && type != DIRENT_TYPE_ROOTDIR &&
+		    size > (guint64)ole->input->size)
+			goto fail;
 
-	dirent = g_new0 (MSOleDirent, 1);
-	dirent->index	     = entry;
-	dirent->size	     = size;
-	dirent->modtime = datetime_from_filetime (ft);
-	/* Store the class id which is 16 byte identifier used by some apps */
-	memcpy(dirent->clsid, data + DIRENT_CLSID, sizeof(dirent->clsid));
+		guint64 ft = GSF_LE_GET_GUINT64 (data + DIRENT_MODIFY_TIME);
 
-	/* root dir is always big block */
-	dirent->use_sb	     = parent && (size < ole->info->threshold);
-	dirent->first_block  = (GSF_LE_GET_GUINT32 (data + DIRENT_FIRSTBLOCK));
-	dirent->is_directory = (type != DIRENT_TYPE_FILE);
-	dirent->children     = NULL;
-	prev  = GSF_LE_GET_GUINT32 (data + DIRENT_PREV);
-	next  = GSF_LE_GET_GUINT32 (data + DIRENT_NEXT);
-	child = GSF_LE_GET_GUINT32 (data + DIRENT_CHILD);
-	name_len = GSF_LE_GET_GUINT16 (data + DIRENT_NAME_LEN);
-	dirent->name = NULL;
-	if (0 < name_len && name_len <= DIRENT_MAX_NAME_SIZE) {
-		gunichar2 uni_name [DIRENT_MAX_NAME_SIZE+1];
-		gchar const *end;
-		int i;
+		MSOleDirent *dirent = g_new0 (MSOleDirent, 1);
+		dirent->index	     = entry;
+		dirent->size	     = size;
+		dirent->modtime = datetime_from_filetime (ft);
+		/* Store the class id which is 16 byte identifier used by some apps */
+		memcpy (dirent->clsid, data + DIRENT_CLSID, sizeof (dirent->clsid));
 
-		/* !#%!@$#^
-		 * Sometimes, rarely, people store the stream name as ascii
-		 * rather than utf16.  Do a validation first just in case.
-		 */
-		if (!g_utf8_validate (data, -1, &end) ||
-		    ((guint8 const *)end - data + 1) != name_len) {
-			/* be wary about endianness */
-			for (i = 0 ; i < name_len ; i += 2)
-				uni_name [i/2] = GSF_LE_GET_GUINT16 (data + i);
-			uni_name [i/2] = 0;
-
-			dirent->name = g_utf16_to_utf8 (uni_name, -1, NULL, NULL, NULL);
-		} else
-			dirent->name = g_strndup ((gchar *)data, (gsize)((guint8 const *)end - data + 1));
-	}
-	/* be really anal in the face of screwups */
-	if (dirent->name == NULL)
-		dirent->name = g_strdup ("");
-	dirent->key = gsf_msole_sorting_key_new (dirent->name);
+		/* root dir is always big block */
+		dirent->use_sb	     = parent && (size < ole->info->threshold);
+		dirent->first_block  = (GSF_LE_GET_GUINT32 (data + DIRENT_FIRSTBLOCK));
+		dirent->is_directory = (type != DIRENT_TYPE_FILE);
+		dirent->children     = NULL;
+		guint32 prev  = GSF_LE_GET_GUINT32 (data + DIRENT_PREV);
+		guint32 next  = GSF_LE_GET_GUINT32 (data + DIRENT_NEXT);
+		guint32 child = GSF_LE_GET_GUINT32 (data + DIRENT_CHILD);
+		guint16 name_len = GSF_LE_GET_GUINT16 (data + DIRENT_NAME_LEN);
+		dirent->name = get_dirent_name ((const char *)data, name_len);
+		/* be really anal in the face of screwups */
+		if (dirent->name == NULL)
+			dirent->name = g_strdup ("");
+		dirent->key = gsf_msole_sorting_key_new (dirent->name);
 
 #if 0
-	printf ("%c '%s' :\tsize = %d\tfirst_block = 0x%x\n",
-		dirent->is_directory ? 'd' : ' ',
-		dirent->name, dirent->size, dirent->first_block);
+		printf ("%c '%s' :\tsize = %d\tfirst_block = 0x%x\n",
+			dirent->is_directory ? 'd' : ' ',
+			dirent->name, dirent->size, dirent->first_block);
 #endif
 
-	if (parent != NULL)
-		parent->children = g_list_insert_sorted (parent->children,
-			dirent, (GCompareFunc)ole_dirent_cmp);
+		if (parent)
+			parent->children = g_list_insert_sorted (parent->children,
+								 dirent, (GCompareFunc)ole_dirent_cmp);
+		else
+			ole->dirent = dirent;
 
-	/* NOTE : These links are a tree, not a linked list */
-	ole_dirent_new (ole, prev, parent, seen_before);
-	ole_dirent_new (ole, next, parent, seen_before);
+		/* NOTE: "prev" and "next" links are a tree, not a linked list */
+		add_todo (todo, parent, prev);
 
-	if (dirent->is_directory)
-		ole_dirent_new (ole, child, dirent, seen_before);
-	else if (child != DIRENT_MAGIC_END)
-		g_warning ("A non directory stream with children ?");
+		if (dirent->is_directory) {
+			add_todo (todo, dirent, child);
+		} else if (child != DIRENT_MAGIC_END)
+			g_warning ("A non directory stream with children?");
 
-	return dirent;
+		add_todo (todo, parent, next);
+	}
+
+	g_array_free (todo, TRUE);
 }
 
 static void
@@ -656,8 +687,8 @@ ole_init_info (GsfInfileMSOle *ole, GError **err)
 	seen_before = g_byte_array_new ();
 	g_byte_array_set_size (seen_before, ((size_t)(ole->bat.num_blocks) << info->bb.shift) / DIRENT_SIZE + 1);
 	memset (seen_before->data, 0, seen_before->len);
-	ole->dirent = info->root_dir =
-		ole_dirent_new (ole, 0, NULL, seen_before);
+	ole_dirent_new (ole, 0, NULL, seen_before);
+	info->root_dir = ole->dirent;
 	g_byte_array_unref (seen_before);
 	if (ole->dirent == NULL) {
 		if (err != NULL)
@@ -968,7 +999,7 @@ GSF_CLASS (GsfInfileMSOle, gsf_infile_msole,
 /**
  * gsf_infile_msole_new:
  * @source: #GsfInput
- * @err: optional place to store an error
+ * @err: (nullable): place to store an error
  *
  * Opens the root directory of an MS OLE file.
  * <note>This adds a reference to @source.</note>
@@ -1002,7 +1033,8 @@ gsf_infile_msole_new (GsfInput *source, GError **err)
 /**
  * gsf_infile_msole_get_class_id:
  * @ole: a #GsfInfileMSOle
- * @res: 16 byte identifier (often a GUID in MS Windows apps)
+ * @res: (out) (array fixed-size=16): identifier (often a GUID in
+ * MS Windows apps)
  *
  * Retrieves the 16 byte indentifier (often a GUID in MS Windows apps)
  * stored within the directory associated with @ole and stores it in @res.
